/**
 *  '$RCSfile: AccessControlList.java,v $'
 *    Purpose: A Class that loads eml-access.xml file containing ACL 
 *             for a metadata document into relational DB
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Jivka Bojilova
 *    Release: @release@
 *
 *   '$Author: bojilova $'
 *     '$Date: 2001/01/20 00:12:22 $'
 * '$Revision: 1.21 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.*;
import java.sql.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.net.URL;
import java.net.MalformedURLException;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.ContentHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;

/** 
 * A Class that loads eml-access.xml file containing ACL for a metadata
 * document into relational DB. It extends DefaultHandler class to handle
 * SAX parsing events when processing the XML stream.
 */
public class AccessControlList extends DefaultHandler {

  static final int ALL = 1;
  static final int WRITE = 2;
  static final int READ = 4;

  private Connection conn;
  private String parserName;
  private Stack elementStack;
  private String server;

  private boolean	processingDTD;
  private String  user;
  private String  group;
  private String  aclid;
  private String 	docname;
  private String 	doctype;
  private String 	systemid;

  private String docurl;
  private Vector resourceURL;
  private Vector resourceID;
  private Vector principal;
  private int    permission;
  private String permType;
  private String permOrder;
  private String publicAcc;
  private String beginTime;
  private String endTime;
  private int    ticketCount;
  private int    serverCode = 1;
  
  /**
   * Construct an instance of the AccessControlList class.
   * It is used by the permission check up from DBQuery and DocumentImpl
   * and for "getaccesscontrol" action
   *
   * @param conn the JDBC connection where acl info is get
   */
  public AccessControlList(Connection conn) throws SQLException
  {
    this.conn = conn;
  }

  /**
   * Construct an instance of the AccessControlList class.
   * It parse acl file and loads acl data into db connection.
   *
   * @param conn the JDBC connection where acl data are loaded
   * @param aclid the Accession# of the document with the acl data
   * @param acl the acl file containing acl data
   * @param user the user connected to MetaCat servlet and owns the document
   * @param group the group to which user belongs
   * @param serverCode the serverid from xml_replication on which this document
   *        resides.
   */
  public AccessControlList(Connection conn, String aclid, Reader acl,
                           String user, String group, int serverCode)
                  throws SAXException, IOException, ClassNotFoundException 
  {
    // Get an instance of the parser
    MetaCatUtil util = new MetaCatUtil();
    String parserName = util.getOption("saxparser");
    this.server = util.getOption("server");

    this.conn = conn;
    this.parserName = parserName;
    this.processingDTD = false;
    this.elementStack = new Stack();
    
    this.user = user;
    this.group = group;
    this.aclid = aclid;
    this.resourceURL = new Vector();
    this.resourceID = new Vector();
    this.principal = new Vector();
    this.permission = 0;
    this.ticketCount = 0;
    this.serverCode = serverCode;
    
    // Initialize the parser and read the queryspec
    XMLReader parser = initializeParser();
    parser.parse(new InputSource(acl));

  }

  /**
   * Construct an instance of the AccessControlList class.
   * It parse acl file and loads acl data into db connection.
   *
   * @param conn the JDBC connection where acl data are loaded
   * @param docid the Accession# of the document with the acl data
   * @param aclfilename the name of acl file containing acl data
   */
  public AccessControlList( Connection conn, String aclid, String aclfilename,
                           String user, String group )
                  throws SAXException, IOException, ClassNotFoundException 
  {
    this(conn, aclid, new FileReader(new File(aclfilename).toString()), 
         user, group, 1);
  }
  
  /* Set up the SAX parser for reading the XML serialized ACL */
  private XMLReader initializeParser() throws SAXException 
  {
    XMLReader parser = null;

    // Get an instance of the parser
    parser = XMLReaderFactory.createXMLReader(parserName);

    // Turn off validation
    parser.setFeature("http://xml.org/sax/features/validation", true);
      
    // Set Handlers in the parser
    // Set the ContentHandler to this instance
    parser.setContentHandler((ContentHandler)this);

    // make a DBEntityResolver instance
    // Set the EntityReslover to DBEntityResolver instance
    EntityResolver eresolver = new DBEntityResolver(conn,this,null);
    parser.setEntityResolver((EntityResolver)eresolver);

    // Set the ErrorHandler to this instance
    parser.setErrorHandler((ErrorHandler)this);

    return parser;
  }
  
  /**
   * callback method used by the SAX Parser when beginning of the document
   */
  public void startDocument() throws SAXException 
  {
    //delete all previously submitted permissions @ relations
    //this happens only on UPDATE of the access file
    try {
      if ( aclid != null ) {
        //first delete all permissions for resources related to @aclid if any
        deletePermissionsForRelatedResources(aclid);
        //then delete all relations with docid of @aclid if any
        deleteRelations(aclid);
      }
    } catch (SQLException sqle) {
      throw new SAXException(sqle);
    }
  }
  
  /**
   * callback method used by the SAX Parser when the start tag of an 
   * element is detected. Used in this context to parse and store
   * the acl information in class variables.
   */
  public void startElement (String uri, String localName, 
                            String qName, Attributes atts) 
         throws SAXException 
  {
    BasicNode currentNode = new BasicNode(localName);
    if (atts != null) {
      int len = atts.getLength();
      for (int i = 0; i < len; i++) {
        currentNode.setAttribute(atts.getLocalName(i), atts.getValue(i));
      }
    }
    if ( currentNode.getTagName().equals("resource") ) {
      permOrder = currentNode.getAttribute("order");
      publicAcc = currentNode.getAttribute("public");
    }
    elementStack.push(currentNode); 
  }

  /**
   * callback method used by the SAX Parser when the text sequences of an 
   * xml stream are detected. Used in this context to parse and store
   * the acl information in class variables.
   */
  public void characters(char ch[], int start, int length)
         throws SAXException 
  {
    String inputString = new String(ch, start, length);
    BasicNode currentNode = (BasicNode)elementStack.peek(); 
    String currentTag = currentNode.getTagName();

    if (currentTag.equals("resourceIdentifier")) {

      // docid of the current resource
      String docid = getDocid(inputString); 
      // URL string of the current resource
      // docurl is declared in the class
      try {
        docurl = (new URL(inputString)).toString();
      } catch (MalformedURLException murle) {
        throw new SAXException(murle.getMessage());
      }
      // collect them in Vector variables
      resourceID.addElement(docid);
      resourceURL.addElement(docurl);
      
      // if it is the local server (originator of the document),
      // check for permission for @user on resource is needed
      // @user must have permission "all" on it(docid)
      if ( serverCode == 1 ) {
        boolean hasPermission = false;
        try {
          hasPermission = hasPermission("ALL",user,docid);
          if ( !hasPermission && group != null ) {
            hasPermission = hasPermission("ALL",group,docid);
          }
        } catch (SQLException e) {
          throw new SAXException(e.getMessage());
        }
        if ( !hasPermission ) {
          throw new SAXException(
          "Permission denied for setting access control on " + docid);
        }
      }
      // end of check for "all" perm on docid

    } else if (currentTag.equals("principal")) {

      principal.addElement(inputString);

    } else if (currentTag.equals("permission")) {

      if ( inputString.trim().toUpperCase().equals("READ") ) {
        permission = permission | READ;
      } else if ( inputString.trim().toUpperCase().equals("WRITE") ) {
        permission = permission | WRITE;
      } else if ( inputString.trim().toUpperCase().equals("ALL") ) {
        permission = permission | ALL;
      } else {
        throw new SAXException("Unknown permission type: " + inputString);
      }

    } else if (currentTag.equals("duration") && 
               beginTime == null && endTime == null ) {
      try {
        beginTime = inputString.substring(0, inputString.indexOf(" "));
        endTime = inputString.substring(inputString.indexOf(" ")+1);
      } catch (StringIndexOutOfBoundsException se) {
        beginTime = inputString;
      }

    } else if (currentTag.equals("ticketCount") && ticketCount == 0 ) {
      try {
        ticketCount = (new Integer(inputString.trim())).intValue();
      } catch (NumberFormatException nfe) {
        throw new SAXException("Wrong integer format for:" + inputString);
      }
    }
  }

  /**
   * callback method used by the SAX Parser when the end tag of an 
   * element is detected. Used in this context to parse and store
   * the acl information in class variables.
   */
  public void endElement (String uri, String localName, String qName)
         throws SAXException 
  {
    BasicNode leaving = (BasicNode)elementStack.pop(); 
    if ( leaving.getTagName().equals("resourceIdentifier") ) {
      
      try {
        // make a relationship for @aclid on the current resource(docurl)
        if ( aclid != null ) {
          insertRelation(aclid, docurl);
        }
      } catch (SQLException sqle) {
        throw new SAXException(sqle);
      }
      
    } else if ( leaving.getTagName().equals("allow") ) {
      
      if ( permission > 0 ) {

        // insert into db calculated permission for the list of principals
        try {
          insertPermissions("allowed");
        } catch (SQLException sqle) {
          throw new SAXException(sqle);
        }
      }

      // reset the allowed permission
      principal = new Vector();
      permission = 0;
      beginTime = null;
      endTime = null;
      ticketCount = 0;
    
    } else if ( leaving.getTagName().equals("deny") ) {
      
      if ( permission > 0 ) {

        // insert into db calculated permission for the list of principals
        try {
          insertPermissions("denied");
        } catch (SQLException sqle) {
          throw new SAXException(sqle);
        }
      }

      // reset the denied permission
      principal = new Vector();
      permission = 0;
      beginTime = null;
      endTime = null;
      ticketCount = 0;

    } else if ( leaving.getTagName().equals("resource") ) {

      // update public access for the list of resources
      try {
        updatePublicAccess(publicAcc);
      } catch (SQLException sqle) {
        throw new SAXException(sqle);
      }
      
      // reset the resource
      resourceID = new Vector();
      resourceURL = new Vector();
      permOrder = null;
      publicAcc = null;
    }

  }

  /** SAX Handler that receives notification of DOCTYPE. Sets the DTD */
  public void startDTD(String name, String publicId, String systemId) 
              throws SAXException {
    docname = name;
    doctype = publicId;
    systemid = systemId;
//    processingDTD = true;
  }

  /** 
   * SAX Handler that receives notification of the start of entities
   */
  public void startEntity(String name) throws SAXException {
    if (name.equals("[dtd]")) {
      processingDTD = true;
    }
  }

  /** 
   * SAX Handler that receives notification of the end of entities
   */
  public void endEntity(String name) throws SAXException {
    if (name.equals("[dtd]")) {
      processingDTD = false;
    }
  }

  /**
   * get the document name
   */
  public String getDocname() {
    return docname;
  }

  /**
   * get the document processing state
   */
  public boolean processingDTD() {
    return processingDTD;
  }
  
  /** Delete from db all permission for resources related to @aclid if any */
  private void deletePermissionsForRelatedResources(String aclid) 
          throws SQLException 
  {
    // delete all acl records for resources related to @aclid if any
    Statement stmt = conn.createStatement();
    stmt.execute("DELETE FROM xml_access WHERE accessfileid = '" + aclid + "'");
    stmt.close();
  }

  /**
   * Deletes all of the relations with a docid of @docid.
   * @param docid the docid to delete.
   */
  public void deleteRelations(String docid)
          throws SQLException 
  {
    Statement stmt = conn.createStatement();
    stmt.execute("DELETE FROM xml_relation WHERE docid='" + docid + "'");
    stmt.close();
  }

  /** Insert relationship into db for @aclid on @resourceURL  */
  private void insertRelation(String aclid, String resourceURL) 
          throws SQLException 
  {
    String aclURL = "metacat://" + server + "/?docid=" + aclid;
    
    // insert relationship 
    PreparedStatement pstmt;
    pstmt = conn.prepareStatement(
            "INSERT INTO xml_relation (docid,subject,relationship,object) " +
            "VALUES (?, ?, ?, ?)");
    pstmt.setString(1, aclid);
    pstmt.setString(2, aclURL);
    pstmt.setString(3, "isaclfor");
    pstmt.setString(4, resourceURL);

    pstmt.execute();
    pstmt.close();
  }

  /** Insert into db calculated permission for the list of principals */
  private void insertPermissions( String permType ) 
          throws SQLException 
  {
    PreparedStatement pstmt;
 
    try {
      pstmt = conn.prepareStatement(
              "INSERT INTO xml_access " + 
              "(docid, principal_name, permission, perm_type, perm_order," +
              "begin_time,end_time,ticket_count, accessfileid) VALUES " +
              "(?,?,?,?,?,to_date(?,'mm/dd/yy'),to_date(?,'mm/dd/yy'),?,?)");
      // Bind the values to the query
      pstmt.setInt(3, permission);
      pstmt.setString(4, permType);
      pstmt.setString(5, permOrder);
      pstmt.setString(6, beginTime);
      pstmt.setString(7, endTime);
      pstmt.setString(9, aclid);
      if ( ticketCount > 0 ) {
        pstmt.setString(8, "" + ticketCount);
      } else {
        pstmt.setString(8, "");
      }
      for ( int i = 0; i < resourceID.size(); i++ ) {
        pstmt.setString(1, (String)resourceID.elementAt(i));
        for ( int j = 0; j < principal.size(); j++ ) {
          pstmt.setString(2, (String)principal.elementAt(j));
          pstmt.execute();
        }
      }
      pstmt.close();

    } catch (SQLException e) {
      throw new 
      SQLException("AccessControlList.insertPermissions(): " + e.getMessage());
    }
  }

  /** Update into db public read access for the list of resources */
  private void updatePublicAccess(String publicAcc) 
          throws SQLException 
  {
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement(
              "UPDATE xml_documents SET public_access = ?" +
              " WHERE docid LIKE ?");
      // Bind the values to the query
      if ( publicAcc == null ) {
        pstmt.setString(1, null);
      } else if ( publicAcc.toUpperCase().equals("YES") ) {
        pstmt.setInt(1, 1);
      } else {
        pstmt.setInt(1, 0);
      }
      for ( int i = 0; i < resourceID.size(); i++ ) {
        pstmt.setString(2, (String)resourceID.elementAt(i));
        pstmt.execute();
      }
      pstmt.close();

    } catch (SQLException e) {
      throw new 
      SQLException("AccessControlList.updatePublicAccess(): " + e.getMessage());
    }
  }

  /** Check for @permission for @principal on @resourceID from db connection */
  public boolean hasPermission ( String permission,
                                 String principal, String resourceID )
                 throws SQLException
  {
    PreparedStatement pstmt;
    // check public access to @resourceID from xml_documents table
    if ( permission.equals("READ") ) {
      try {
        pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " +
                "WHERE docid LIKE ? AND public_access = 1");
        // Bind the values to the query
        pstmt.setString(1, resourceID);

        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean hasRow = rs.next();
        pstmt.close();
        if (hasRow) {
          return true;
        }
//System.out.println("Passed the check for public access");      

      } catch (SQLException e) {
        throw new 
        SQLException("AccessControlList.hasPermission(): " +
                     "Error checking document's public access. " +
                      e.getMessage());
      }
    }
    
    // since owner of resource has all permission on it,
    // check if @principal is owner of @resourceID in xml_documents table
    if ( principal != null ) {
      try {
        pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " +
                "WHERE docid LIKE ? AND user_owner LIKE ?");
        // Bind the values to the query
        pstmt.setString(1, resourceID);
        pstmt.setString(2, principal);

        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean hasRow = rs.next();
        pstmt.close();
        if (hasRow) {
          return true;
        }
//System.out.println("Passed the check for ownership");      
     
      } catch (SQLException e) {
        throw new 
        SQLException("AccessControlList.hasPermission(): " +
                     "Error checking document's ownership. " + e.getMessage());
      }

      // check @principal's @permission on @resourceID from xml_access table
      int accessValue = 0;
      int ticketCount = 0;
      String permOrder = "";
      try {
        pstmt = conn.prepareStatement(
                "SELECT permission, perm_order, ticket_count " +
                "FROM xml_access " +
                "WHERE docid LIKE ? " + 
                "AND principal_name LIKE ? " +
                "AND sysdate BETWEEN nvl(begin_time,sysdate) " +
                                "AND nvl(end_time,sysdate) " +
                "AND perm_type LIKE ?");
        // check if it is "denied" first
        // Bind the values to the query
        pstmt.setString(1, resourceID);
        pstmt.setString(2, principal);
        pstmt.setString(3, "denied");

        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean hasRows = rs.next();
        while ( hasRows ) {
          accessValue = rs.getInt(1);
          permOrder = rs.getString(2);
          ticketCount = rs.getInt(3);
          if ( ( accessValue & intValue(permission) ) == intValue(permission) &&
               ( permOrder.equals("allowFirst") ) &&
               ( rs.wasNull() || ticketCount > 0 ) ) {
            if ( !rs.wasNull() && ticketCount > 0 ) {
              decreaseNumberOfAccess(accessValue,principal,resourceID,"denied");
            }
            pstmt.close();
            return false;
          }
          hasRows = rs.next();
        }
//System.out.println("Passed the check for denied access");      

        // it is not denied then check if it is "allowed"
        // Bind the values to the query
        pstmt.setString(1, resourceID);
        pstmt.setString(2, principal);
        pstmt.setString(3, "allowed");

        pstmt.execute();
        rs = pstmt.getResultSet();
        hasRows = rs.next();
        while ( hasRows ) {
          accessValue = rs.getInt(1);
          ticketCount = rs.getInt(3);
          if ( ( accessValue & intValue(permission) )==intValue(permission) &&
               ( rs.wasNull() || ticketCount > 0 ) ) {
            if ( !rs.wasNull() && ticketCount > 0 ) {
              decreaseNumberOfAccess(accessValue,principal,resourceID,"allowed");
            }
            pstmt.close();
            return true;
          }
          hasRows = rs.next();
        }
//System.out.println("Passed the check for allowed access");      

        // ???
        // here there could be a check for the group's permission like 
        // selfrecursive call to hasPermission(conn,permission,group,recourceId)
        //
      
        pstmt.close();
        return false;
  
      } catch (SQLException e) {
        throw new 
        SQLException("AccessControlList.hasPermission(): " +
                     "Error checking document's permission. " + e.getMessage());
      }
    }
    
    return false;
  }

  /** decrease the number of access to @resourceID for @principal */
  private void decreaseNumberOfAccess(int permission, String principal,
                                      String resourceID, String permType) 
               throws SQLException
  {
    PreparedStatement pstmt;
    pstmt = conn.prepareStatement(
            "UPDATE xml_access SET ticket_count = ticket_count - 1 " +
            "WHERE docid LIKE ? " +
            "AND principal_name LIKE ? " +
            "AND permission LIKE ? " +
            "AND sysdate BETWEEN nvl(begin_time,sysdate) " +
                            "AND nvl(end_time,sysdate) " +
            "AND perm_type LIKE ?");
    // Bind the values to the query
    pstmt.setString(1, resourceID);
    pstmt.setString(2, principal);
    pstmt.setInt(3, permission);
    pstmt.setString(4, permType);

    pstmt.execute();
    pstmt.close();
  }
 
  // get the int value of READ, WRITE or ALL
  private int intValue ( String permission )
  {
    if ( permission.equals("READ") ) {
      return READ;
    } else if ( permission.equals("WRITE") ) {
      return WRITE;
    } else if ( permission.equals("ALL") ) {
      return ALL;
    }
    
    return -1;
  }
  
  // get docid from @url
  private String getDocid ( String url ) throws SAXException
  {
    MetaCatUtil util = new MetaCatUtil();
    try {
      URL urlobj = new URL(url);
      Hashtable urlParams = util.parseQuery(urlobj.getQuery());
      if ( urlParams.containsKey("docid") ) {
        return (String)urlParams.get("docid"); // return the docid value
      } else {
        throw new 
        SAXException("\"docid\" not specified within " + url);
      }
    } catch (MalformedURLException e) {
      throw new SAXException(e.getMessage());
    }
  }  

}
