/**
 *  '$RCSfile: AccessionNumber.java,v $'
 *    Purpose: A class that gets Accession Number, check for uniqueness
 *             and register it into db
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Jivka Bojilova, Matt Jones
 *    Release: @release@
 *
 *   '$Author: jones $'
 *     '$Date: 2001/01/18 19:51:57 $'
 * '$Revision: 1.18 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.net.*;
import java.sql.*;

/**
 * (on insert of XML document)
 * Generates a unique Accession Number or if provided check it 
 * for uniqueness and register it into the db connection 
 * (on update or delete of XML document)
 * Check for existance of provided Accession Number
 * 
 */
public class AccessionNumber  {
  
  private Connection conn = null;
  private String sitecode = null;
  private String sep = null;
    
  /** 
   * Construct an AccessionNumber 
   */
  public AccessionNumber () 
         throws SQLException, ClassNotFoundException  {
        
    MetaCatUtil util = new MetaCatUtil();

    this.sitecode = util.getOption("sitecode");
    this.sep = util.getOption("accNumSeparator");

    // Open a connection to the database
    this.conn = util.openDBConnection();
  }  

  /** 
   * Construct an AccessionNumber
   *
   * @param conn the db connection to read from and write Accession# to
   */
  public AccessionNumber (Connection conn) {
        
    MetaCatUtil util = new MetaCatUtil();

    this.sitecode = util.getOption("sitecode");
    this.sep = util.getOption("accNumSeparator");
    this.conn = conn;

  }

  /**
   * Generate an Accession Number of form <sidecode>.<createdate>
   */
  public String generate(String accNumber, String action) 
                throws AccessionNumberException, SQLException 
  {
    try {
      // INSERT
      if ( action.equals("INSERT")) {

        // get a new Accession#
        if ( accNumber == null ) {
          String sitecode = getSitecode();
          String uniqueid = getUniqueID();

          return sitecode + sep + uniqueid;

        // Accession# is not used; return it
        } else if ( !accNumberUsed(accNumber) ) {
          return accNumber;

        // Accession# is used; throw an exception to prevent the insertion
        } else {
          throw new AccessionNumberException
                    ("Accession number " + accNumber + " is already in use.");
        }

      // UPDATE or DELETE
      } else if ( action.equals("UPDATE") || action.equals("DELETE")) {

        // Accession# is not provided; throw an exception to prevent the action
        if ( accNumber == null ) {
          throw new AccessionNumberException("Accession number is required.");

        // Accession# is not current (not in xml_documents); throw an exception
        } else if ( !accNumberIsCurrent(accNumber) ) {
          throw new AccessionNumberException
          ("Document not found for Accession number " + accNumber);

        // Accession# is current (present in xml_documents); return it
        } else {
          return accNumber;
        }
      }
 
    } catch (SQLException e) {
      throw new SQLException
      ("Error on AccessionNumber.generate(accNumber, action): "+e.getMessage());
    }    
 
    // never comes here
    throw new
    AccessionNumberException("Fatal Error in accession number generation.");
  }
  
  // get local sitecode
  private String getSitecode()
  {
    return this.sitecode;
  }

  // get Unique ID from DB sequence
  private String getUniqueID () throws SQLException
  {
    String uniqueid;
    
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement
              ("SELECT accnum_uniqueid_seq.nextval FROM dual");
      pstmt.execute();

      ResultSet rs = pstmt.getResultSet();
      boolean hasRow = rs.next();
      uniqueid = rs.getString(1);
      pstmt.close();
      
    } catch (SQLException e) {
      throw new 
      SQLException("Error on AccessionNumber.getUniqueID(): "+e.getMessage());
    }

    return uniqueid;
  }

  /** check for existence of Accesssion Number xml_acc_numbers table */
  public boolean accNumberUsed ( String accNumber )
                  throws SQLException {
        
    boolean hasAccNumber = false;
        
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " + 
                "WHERE docid LIKE ? " +
                "UNION " +
                "SELECT 'x' FROM xml_revisions " +
                "WHERE docid LIKE ?");
      pstmt.setString(1,accNumber);
      pstmt.setString(2,accNumber);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      hasAccNumber = rs.next();
      pstmt.close();
            
    } catch (SQLException e) {
      throw new SQLException
      ("Error on AccessionNumber.accNumberUsed(accNumber): " + e.getMessage());
    }    
        
    return hasAccNumber;
  }    
    
  /** check for existence of Accesssion Number in xml_documents table */
  private boolean accNumberIsCurrent(String accNumber) throws SQLException {
        
    boolean hasCurrentAccNumber = false;
        
    try {
      PreparedStatement pstmt;
      pstmt = conn.prepareStatement(
                "SELECT 'x' FROM xml_documents " + 
                "WHERE docid LIKE ?");
      pstmt.setString(1, accNumber);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      hasCurrentAccNumber = rs.next();
      pstmt.close();
            
    } catch (SQLException e) {
      throw new SQLException(
          "Error on AccessionNumber.accNumberIsCurrent(String accNumber): " +
          e.getMessage());
    }    
  
    return hasCurrentAccNumber;
  }    

}
